﻿
// **************************************************************************
//    Class Created by Mick Doherty (Dotnetrix) July 2010
//    http://dotnetrix.co.uk/
//
//    ...for GlassUI (an AeroGlass UI Library)
//    http://glassui.codeplex.com/
// **************************************************************************

using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using GlassUI.Design;

namespace GlassUI
{
    [DefaultProperty("Checked")]
    [DefaultBindingProperty("CheckState")]
    [DefaultEvent("CheckedChanged")]
    [ToolboxBitmap(typeof(System.Windows.Forms.CheckBox))]
    public class GlassCheckBox : GlassAutoSizeControlBase
    {

        public GlassCheckBox()
            : base()
        {
            SetStyle(ControlStyles.SupportsTransparentBackColor | ControlStyles.AllPaintingInWmPaint | ControlStyles.OptimizedDoubleBuffer | ControlStyles.ResizeRedraw, true);
            SetStyle(ControlStyles.Selectable | ControlStyles.StandardClick, true);
            SetStyle(ControlStyles.StandardDoubleClick, false);
        }

        #region Event Declarations

        public event EventHandler CheckedChanged;
        public event EventHandler CheckStateChanged;

        #endregion

        #region Private Instance Variables

        private Appearance appearance = Appearance.Normal;

        private Color checkColor = Color.Blue;

        private CheckState checkState = CheckState.Unchecked;
        private bool autoCheck = true;
        private bool threeState;

        private ControlState controlState = ControlState.Normal;

        #endregion

        #region Public Properties

        [DefaultValue(typeof(Appearance), "Normal")]
        [GlassUIDescription("CheckBoxAppearanceDescr", "System.Windows.Forms")]
        public Appearance Appearance
        {
            get { return appearance; }
            set
            {
                appearance = value;
                if (AutoSize)
                    this.Size = this.GetPreferredSize(this.Size);
            }
        }

        [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
        public override Image BackgroundImage
        {
            get
            {
                return base.BackgroundImage;
            }
            set
            {
                base.BackgroundImage = value;
            }
        }

        [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
        public override ImageLayout BackgroundImageLayout
        {
            get
            {
                return base.BackgroundImageLayout;
            }
            set
            {
                base.BackgroundImageLayout = value;
            }
        }

        [DefaultValue(typeof(Color), "Blue")]
        [GlassUIDescription("CheckBoxCheckColorDescription")]
        public Color CheckColor
        {
            get { return checkColor; }
            set
            {
                if (!checkColor.Equals(value))
                {
                    checkColor = value;
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(false)]
        [GlassUIDescription("CheckBoxCheckedDescr", "System.Windows.Forms")]
        public bool Checked
        {
            get { return checkState != CheckState.Unchecked; }
            set
            {
                if (value != this.Checked)
                {
                    this.CheckState = value ? CheckState.Checked : CheckState.Unchecked;
                }
                this.OnCheckedChanged(EventArgs.Empty);
                this.Invalidate();
            }
        }

        [DefaultValue(typeof(CheckState), "Unchecked")]
        [GlassUIDescription("CheckBoxCheckStateDescr", "System.Windows.Forms")]
        public CheckState CheckState
        {
            get { return checkState; }
            set
            {
                if (!Enum.IsDefined(typeof(CheckState), value))
                    throw new InvalidEnumArgumentException("value", (int)value, typeof(CheckState));
                if (checkState != value)
                {
                    this.checkState = value;
                    this.OnCheckStateChanged(EventArgs.Empty);
                    this.Invalidate();
                }
            }
        }

        [DefaultValue(true)]
        [GlassUIDescription("CheckBoxAutoCheckDescr", "System.Windows.Forms")]
        public bool AutoCheck
        {
            get { return autoCheck; }
            set { autoCheck = value; }
        }

        [DefaultValue(false)]
        [GlassUIDescription("CheckBoxThreeStateDescr", "System.Windows.Forms")]
        public bool ThreeState
        {
            get { return threeState; }
            set { threeState = value; }
        }

        internal ControlState ControlState
        {
            get { return controlState; }
            set
            {
                if (controlState != value)
                {
                    controlState = value;
                    this.Invalidate();
                }
            }
        }

        #endregion

        #region Overridden Properties

        protected override Size DefaultSize
        {
            get
            {
                return new Size(104, 24);
            }
        }

        #endregion

        #region Overridden Methods

        public override Size GetPreferredSize(Size proposedSize)
        {
            int h = this.Appearance == Appearance.Button ? 0 : this.FontHeight;

            Size checkSize = new Size(h, h);
            Size availableTextSize = Size.Empty;
            Size textSize = Size.Empty;

            Rectangle contentBounds = Rectangle.Empty;
            TextFormatFlags format = TextFormatFlags.TextBoxControl | TextFormatFlags.SingleLine | TextFormatFlags.NoPadding;

            availableTextSize = new Size(proposedSize.Width - checkSize.Width, proposedSize.Height);

            if (this.Text.Length > 0)
            {
                using (Graphics graphics = this.CreateGraphics())
                {
                    textSize = GlassUtilities.MeasureThemedText(graphics, this.Text, this.Font, availableTextSize, format, 8);
                }
            }

            contentBounds.Width = checkSize.Width + textSize.Width + 4;
            contentBounds.Height = Math.Max(checkSize.Height, textSize.Height);

            contentBounds.Width += Padding.Horizontal;
            contentBounds.Height += Padding.Vertical;

            return contentBounds.Size;
        }

        protected override void OnPaintBackground(PaintEventArgs pevent)
        {
            base.OnPaintBackground(pevent);

            pevent.Graphics.SmoothingMode = SmoothingMode.HighQuality;

            int h = this.FontHeight;

            Rectangle r = this.Appearance == Appearance.Button ? this.ClientRectangle : new Rectangle(2, (this.Height / 2) - (h / 2), h, h);

            if (this.BackColor.A < 255)
                DrawParentBackground(pevent.Graphics);

            if (this.Appearance == Appearance.Button)
            {
                pevent.Graphics.CompositingQuality = CompositingQuality.HighQuality;
                using (GraphicsPath path = GlassUtilities.RoundedRectangle(r, this.CornerRadius, this.RoundedCorners))
                {
                    pevent.Graphics.SetClip(path, CombineMode.Replace);
                    ControlState state = this.ControlState;
                    if (this.Checked)
                        state = ControlState.Pressed;
                    GlassUtilities.DrawGlowGradient(pevent.Graphics, r, this.Checked ? this.CheckColor : this.BackColor, this.GlowColor, state);
                    pevent.Graphics.ResetClip();
                    GlassUtilities.DrawRoundedBorder(pevent.Graphics, r, this.CornerRadius, this.RoundedCorners, state == ControlState.Pressed ? this.GlowColor : Color.Empty);
                }
            }
            else
            {
                using (LinearGradientBrush brush = new LinearGradientBrush(r, Color.Empty, this.BackColor, LinearGradientMode.ForwardDiagonal))
                {
                    using (Pen pen = new Pen(brush))
                    {
                        Color darkColor = Color.FromArgb(192, Color.Black);

                        /* Draw Glow */
                        if (this.controlState == ControlState.Hot)
                        {
                            Rectangle glowBounds = this.ClientRectangle;
                            using (PathGradientBrush glowBrush = new PathGradientBrush(GlassUtilities.RoundedRectangle(glowBounds, 2)))
                            {
                                glowBrush.CenterColor = this.GlowColor;
                                glowBrush.SurroundColors = new Color[] { Color.Transparent };
                                pevent.Graphics.FillRectangle(glowBrush, glowBounds);
                            }
                        }

                        /* Draw Dark Border */
                        pen.Color = darkColor;
                        pevent.Graphics.DrawRectangle(pen, r);

                        /* Draw Light Border */
                        r.Inflate(-1, -1);
                        Color lightColor = Color.FromArgb(192, Color.White);
                        pen.Color = this.ControlState == ControlState.Hot ? this.GlowColor : lightColor;
                        pevent.Graphics.DrawRectangle(pen, r);

                        /* Draw Gradient Border */
                        r.Inflate(-1, -1);
                        brush.LinearColors = new Color[] { darkColor, this.BackColor };
                        pen.Brush = brush;
                        pevent.Graphics.DrawRectangle(pen, r);

                        /* Draw Gradient Fill */
                        brush.LinearColors = new Color[] { darkColor, lightColor };
                        pevent.Graphics.FillRectangle(brush, r);

                        /* Draw Check Mark */
                        if (this.Checked)
                        {
                            brush.LinearColors = new Color[] { checkColor, checkColor };
                            if (CheckState == CheckState.Checked)
                            {
                                Point[] pt = new Point[]{
                                        new Point(r.Left,r.Top + r.Height/2),
                                        new Point(r.Left + r.Width/3,r.Top + r.Height),
                                        new Point(r.Left + r.Width,r.Top),
                                        new Point(r.Left+r.Width/3, r.Top + (int)(r.Height *0.7)),
                                        new Point(r.Left,r.Top + r.Height/2)};
                                using (GraphicsPath p = new GraphicsPath())
                                {
                                    p.AddLines(pt);
                                    p.CloseFigure();
                                    pevent.Graphics.FillPath(brush, p);
                                }
                            }
                            else
                            {
                                using (GraphicsPath path = GlassUtilities.RoundedRectangle(r, 2))
                                {
                                    brush.LinearColors = new Color[] {lightColor, checkColor };
                                    pevent.Graphics.FillPath(brush, path);
                                }
                            }
                        }

                    }
                }
            }

            //Draw Text
            DrawText(pevent.Graphics);

            //Draw Focus
            if (this.Focused && this.ShowFocusCues)
                ControlPaint.DrawFocusRectangle(pevent.Graphics, this.ClientRectangle, this.ForeColor, this.BackColor);

        }

        protected override void OnClick(EventArgs e)
        {
            if (this.AutoCheck)
            {
                switch (checkState)
                {
                    case CheckState.Indeterminate:
                        CheckState = 0;
                        break;
                    case CheckState.Checked:
                        if (this.ThreeState)
                            CheckState++;
                        else
                            CheckState = 0;
                        break;
                    default:
                        CheckState++;
                        break;
                }
            }
            base.OnClick(e);
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            base.OnKeyDown(e);
            if (e.KeyCode == Keys.Space)
            {
                this.ControlState = ControlState.Pressed;
            }
            this.OnControlStateChanged(EventArgs.Empty);
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            base.OnKeyUp(e);
            if (e.KeyCode == Keys.Space)
            {
                if (this.ControlState == ControlState.Pressed)
                {
                    this.PerformClick();
                }
                this.ControlState = ControlState.Focused;
            }
            this.OnControlStateChanged(EventArgs.Empty);
        }

        protected override void OnMouseEnter(EventArgs e)
        {
            base.OnMouseEnter(e);
            this.ControlState = ControlState.Hot;
        }

        protected override void OnMouseLeave(EventArgs e)
        {
            base.OnMouseLeave(e);
            this.ControlState = ControlState.Normal;
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);
            if (e.Button == MouseButtons.Left)
            {
                this.Focus();
                this.ControlState = ControlState.Pressed;
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);
            this.ControlState = ControlState.Focused;
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);
            if (new Rectangle(Point.Empty, this.Size).Contains(e.X, e.Y) && e.Button == MouseButtons.Left)
                this.ControlState = ControlState.Pressed;
            else
            {
                this.ControlState = ControlState.Hot;
            }
        }

        protected override void OnGotFocus(EventArgs e)
        {
            base.OnGotFocus(e);
            this.ControlState = ControlState.Focused;
        }

        protected override void OnLostFocus(EventArgs e)
        {
            base.OnLostFocus(e);
            this.ControlState = ControlState.Normal;
        }

        protected override void OnEnabledChanged(EventArgs e)
        {
            base.OnEnabledChanged(e);
            if (this.Enabled)
                this.ControlState = ControlState.Normal;
            else
                this.ControlState = ControlState.Disabled;
        }

        public override string ToString()
        {
            String val = base.ToString();
            return val + ", CheckState: " + (int)this.CheckState;
        }

        #endregion

        public void PerformClick()
        {
            if (this.CanSelect)
                this.OnClick(EventArgs.Empty);
        }

        protected virtual void OnCheckedChanged(EventArgs eventArgs)
        {
            if (this.CheckedChanged != null)
                this.CheckedChanged(this, eventArgs);
        }

        protected virtual void OnCheckStateChanged(EventArgs eventArgs)
        {
            if (this.CheckStateChanged != null)
                this.CheckStateChanged(this, eventArgs);
        }

        protected virtual void OnControlStateChanged(EventArgs eventArgs)
        {
            if (this.CheckStateChanged != null)
                this.CheckStateChanged(this, eventArgs);
        }

        private void DrawText(Graphics g)
        {
            Color textColor = this.ForeColor;
            
            bool isButton = this.Appearance == Appearance.Button;
            int h = isButton ? 0 : this.FontHeight;
            
            Rectangle bounds = Rectangle.FromLTRB(isButton ? 0 : h + 4, 0, this.Width, this.Height);
            
            if (!this.Enabled)
                textColor = SystemColors.GrayText;

            TextFormatFlags format = TextFormatFlags.EndEllipsis | TextFormatFlags.NoPadding | TextFormatFlags.SingleLine | TextFormatFlags.VerticalCenter;
            if (!this.ShowKeyboardCues)
                format |= TextFormatFlags.HidePrefix;

            GlassUtilities.DrawThemedText(g, this.Text, this.Font, bounds, textColor, format, 8);

        }

    }
}
